/*
 * decodeur de 8 aiguillages sur Arduino UNO/NANO
 * Les aiguillages sont commandes par des sorties sur les broches a partir de la broche 3 par paire (aiguillage 1 =  (3,4) etc...)
 * ces sorties peuvent etre utilisees directement pour commander des ULN2803 ou les inputs de relais ou L298N 
 * 
 */

/******************************************************************************************
 * 
 * LES PARAMETRES DE CONFIGURATION a definir par l'utilisateur
 * CONSOLE : pour avoir le deroulement du programme, si pas necessaire supprimer la ligne
 * FIRST_ID_DCC : premiere adresse DCC de depart, premier aiguillage (FIRST_ID_DCC sur broche 3), second aiguillage (FIRST_ID_DCC+1 sur broche 4, etc..)
 * NB_TURNOUT : le nombre d'aiguillages = 8
 * 
 *********************************************************************************************/
 
 #define CONSOLE                         // sortie console du deroulement
 #define DEBIT_CONSOLE    115200         // le debit de la console
 #define FIRST_ID_DCC      40            // ID DCC de départ

 /***********************************************************************************************
  * 
  * LES PARAMETRES INTERNES
  * 
  * ************************************************************************************************/
 
//////////////////// LE DCC  /////////////////////////////////////////////////////////////////

#include <DCC_Decoder.h>                 // la bibliothèque de Minabay
#define kDCC_INTERRUPT    0             // la pin 2 qui recoit les interruptions du signal DCC
int adresse_precedente = 0;             // evite les commandes multiples DCC
int ancienne_position = 2;
volatile boolean MAJ_activation;        // si des MAJ a faire apres controle du DCC

/////////////////// LES AIGUILLAGES  ///////////////////////////////////////////////////////

#define FIRST_PIN         3             // broche de depart des aiguillages
#define BOBINE1           0             // adresse DCC/0
#define BOBINE2           1             // adresse DCC/1 
#define ACTIVATION        250           // duree d'activation des bobines en ms ( amodifier si necessaire)
#define NB_TURNOUT        8             // le nombre  d'aiguillages 

//// STRUCTURE D'UN AIGUILLAGE ////

struct TurnOut {
  int adresse;                    // son adresse DCC
  int position_actuelle;          // directe / deviee
  int bobine1;                    // broche bobine 1
  int bobine2;                    // broche bobine 2
  boolean demande_activation;     // demande d'activation
};
TurnOut turnout[NB_TURNOUT];      // l'ensemble des aiguillages

/********************************************************************
 * 
 * METHODE APPELEE SI UNE DEMANDE DE MISE A JOUR PAR LE DCC
 * Determine periodiquemùent les bobines a activer
 * 
 *******************************************************************/
 
void activation_aiguillage() {
 
  for (int i = 0; i < NB_TURNOUT; i++)                 // on balaie tous les aiguillages
  {
    if (turnout[i].demande_activation == true)         // si l'aiguillage i est en attente d'activation
    {
      switch(turnout[i].position_actuelle)            // on regarde la position demandee
      { 
        case BOBINE1 :{                                           // si bobine 1
                      digitalWrite(turnout[i].bobine1,HIGH);     // activation
                      delay(ACTIVATION);                         // duree de l'activation
                      digitalWrite(turnout[i].bobine1,LOW);      // arret de l'activation
                      #ifdef CONSOLE
                        Serial.print("activation -> aiguillage");Serial.print(i);Serial.println(" : bobine1");
                      #endif
                      break;}
        case BOBINE2 : {                                          // si bobine 2
                      digitalWrite(turnout[i].bobine2,HIGH);      // activation
                      delay(ACTIVATION);                          // duree de l'activation
                      digitalWrite(turnout[i].bobine2,LOW);       // arret de l'activation
                      #ifdef CONSOLE
                        Serial.print("activation -> aiguillage");Serial.print(i);Serial.println(" : bobine2");
                      #endif
                     break;}
           }
       turnout[i].demande_activation = false;                // l'aiguillage a etait traite
      }
    }
  MAJ_activation = false;                                   // on a fait toutes les MAJ
}

/************************************************************************************* 
 *  
 *  METHODE APPELEE A CHAQUE IT DU DCC
 *  determine si un aiguillage est concerne par la commande DCC
 *  
 ***********************************************************************************/

void BasicAccDecoderPacket_Handler(int address, boolean activate, byte data)
{  
  address -= 1; address *= 4; address += 1; address += (data & 0x06) >> 1;    // decodage adresse DCC
  int bobine = (data & 0x01) ? BOBINE1 : BOBINE2;             // quelle bobine est concernee DCC/bobine
  // si on change d'adresse ou de position
  if ((address != adresse_precedente) | ((bobine != ancienne_position)&(address == adresse_precedente))){
    if ((address >= FIRST_ID_DCC) && (address < FIRST_ID_DCC + NB_TURNOUT))  // si une de nos adresses
    {turnout[address - FIRST_ID_DCC].demande_activation =  true;            // aiguillage en demande d'activation
    turnout[address - FIRST_ID_DCC].position_actuelle = bobine;            // etat demande (n° de bobine)
    MAJ_activation = true;                                                // MAJ demandee 
    }
   adresse_precedente = address; ancienne_position = bobine;            // evite de traiter plusieurs fois la meme commande DCC
  }
}
/**********************************************************************************************
 * 
 * le setup
 * initialise le DCC et les aiguillages
 * 
 ******************************************************************************************/

void setup() {
  int decal = 0;
  for (int i=0; i<NB_TURNOUT; i++){              // pour tous les aiguillages
    turnout[i].adresse = FIRST_ID_DCC + i;          // son ID DCC
    turnout[i].demande_activation = false;          // pas de demande d'activation
    turnout[i].bobine1 = decal+FIRST_PIN;           // numero de broche de la bobine 1
    pinMode(turnout[i].bobine1, OUTPUT);            // bobine 1   en sortie (ID DCC/0)
    digitalWrite(turnout[i].bobine1, LOW);          // bobine 1 au repos
    turnout[i].bobine2 = 1+decal+FIRST_PIN;         // numero de la broche de la bobine 2
    pinMode(turnout[i].bobine2, OUTPUT);            // bobine 2  en sortie  (ID DCC/1)
    digitalWrite(turnout[i].bobine2, LOW);          // bobine 2 au repos
    decal+=2;                                   // on passe aux aiguillages suivants  
  }
  DCC.SetBasicAccessoryDecoderPacketHandler(BasicAccDecoderPacket_Handler, true);     // le DCC
  DCC.SetupDecoder( 0x00, 0x00, kDCC_INTERRUPT );                                     // son IT
  MAJ_activation = false;                                  // pas de MAJ a priori
  #ifdef CONSOLE
   Serial.begin(DEBIT_CONSOLE);
   for (int i=0; i<NB_TURNOUT; i++){
    Serial.print("aiguillage");Serial.print(i);Serial.print(" , code DCC : ");Serial.println(turnout[i].adresse); 
    Serial.print("\t bobine 1 sur broche : ");Serial.println(turnout[i].bobine1);
    Serial.print("\t bobine 2 sur broche : ");Serial.println(turnout[i].bobine2);
  }
  #endif
}

/*************************************************************************
 * boucle
 * 
 ******************************************************************/
 
void loop() {
  DCC.loop();                                             // demande DCC ?
  if (MAJ_activation) {activation_aiguillage();}          // si oui activation des aiguillages
}

